//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/BeamItems.cpp
//! @brief     Implements BeamItem hierarchy
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/BeamItems.h"
#include "Base/Axis/IAxis.h"
#include "Base/Const/Units.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "GUI/Model/Device/BeamAngleItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/Model/Device/FootprintItems.h"
#include "GUI/Model/Device/PointwiseAxisItem.h"
#include <cmath>

namespace {

namespace Tag {

const QString Intensity("Intensity");
const QString Wavelength("Wavelength");
const QString AzimuthalAngle("AzimuthalAngle");
const QString InclinationAngle("InclinationAngle");
const QString Footprint("Footprint");
const QString BaseData("BaseData");
const QString ExpandBeamParametersGroupbox("ExpandBeamParametersGroupbox");
const QString ExpandFootprintGroupbox("ExpandFootprintGroupbox");

} // namespace Tag

} // namespace

//  ************************************************************************************************
//  BeamItem
//  ************************************************************************************************

BeamItem::BeamItem()
{
    m_intensity.init("Intensity", "Beam intensity in neutrons (or gammas) per sec.", 1e+08,
                     Unit::unitless, 3 /* decimals */, RealLimits::limited(0.0, 1e+32),
                     "intensity");

    m_azimuthalAngleItem.reset(new BeamAzimuthalAngleItem());
}

void BeamItem::writeTo(QXmlStreamWriter* w) const
{
    ASSERT(m_wavelengthItem);
    ASSERT(m_inclinationAngleItem);

    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // intensity
    w->writeStartElement(Tag::Intensity);
    m_intensity.writeTo(w);
    w->writeEndElement();

    // wavelength
    w->writeStartElement(Tag::Wavelength);
    m_wavelengthItem->writeTo(w);
    w->writeEndElement();

    // azimuthal angle
    w->writeStartElement(Tag::AzimuthalAngle);
    m_azimuthalAngleItem->writeTo(w);
    w->writeEndElement();

    // inclination angle
    w->writeStartElement(Tag::InclinationAngle);
    m_inclinationAngleItem->writeTo(w);
    w->writeEndElement();

    // beam parameters groupbox: is expanded?
    w->writeStartElement(Tag::ExpandBeamParametersGroupbox);
    XML::writeAttribute(w, XML::Attrib::value, m_expandBeamParameters);
    w->writeEndElement();
}

void BeamItem::readFrom(QXmlStreamReader* r)
{
    ASSERT(m_wavelengthItem);
    ASSERT(m_inclinationAngleItem);

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // intensity
        if (tag == Tag::Intensity) {
            m_intensity.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // wavelength
        } else if (tag == Tag::Wavelength) {
            m_wavelengthItem->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // azimuthal angle
        } else if (tag == Tag::AzimuthalAngle) {
            m_azimuthalAngleItem->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // inclination angle
        } else if (tag == Tag::InclinationAngle) {
            m_inclinationAngleItem->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // beam parameters groupbox: is expanded?
        } else if (tag == Tag::ExpandBeamParametersGroupbox) {
            XML::readAttribute(r, XML::Attrib::value, &m_expandBeamParameters);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

double BeamItem::wavelength() const
{
    ASSERT(m_wavelengthItem);
    return m_wavelengthItem->wavelength();
}

void BeamItem::setWavelength(double value)
{
    ASSERT(m_wavelengthItem);
    m_wavelengthItem->resetToValue(value);
}

BeamWavelengthItem* BeamItem::wavelengthItem() const
{
    ASSERT(m_wavelengthItem);
    return m_wavelengthItem.get();
}

void BeamItem::setInclinationAngle(double value)
{
    ASSERT(m_inclinationAngleItem);
    m_inclinationAngleItem->resetToValue(value);
}

BeamDistributionItem* BeamItem::inclinationAngleItem() const
{
    ASSERT(m_inclinationAngleItem);
    return m_inclinationAngleItem.get();
}

double BeamItem::azimuthalAngle() const
{
    ASSERT(m_azimuthalAngleItem);
    return m_azimuthalAngleItem->azimuthalAngle();
}

void BeamItem::setAzimuthalAngle(double value)
{
    ASSERT(m_azimuthalAngleItem);
    m_azimuthalAngleItem->resetToValue(value);
}

BeamAzimuthalAngleItem* BeamItem::azimuthalAngleItem() const
{
    ASSERT(m_azimuthalAngleItem);
    return m_azimuthalAngleItem.get();
}

std::shared_ptr<Beam> BeamItem::createBeam() const
{
    double lambda = wavelength();
    double inclination_angle = Units::deg2rad(getInclinationAngle());
    double azimuthal_angle = Units::deg2rad(azimuthalAngle());

    auto result =
        std::make_shared<Beam>(InBeam(intensity(), lambda, inclination_angle, azimuthal_angle));

    return result;
}

//  ************************************************************************************************
//  SpecularBeamIitem
//  ************************************************************************************************

ScanItem::ScanItem(const InstrumentItem* owningInstrument)
{
    m_inclinationAngleItem.reset(new SpecularBeamInclinationItem(owningInstrument));
    m_wavelengthItem.reset(new SpecularBeamWavelengthItem);

    m_footprint.init("Type", "Footprint type");
}

void ScanItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    BeamItem::writeTo(w);
    w->writeEndElement();

    // footprint
    w->writeStartElement(Tag::Footprint);
    m_footprint.writeTo(w);
    w->writeEndElement();

    // footprint groupbox: is expanded?
    w->writeStartElement(Tag::ExpandFootprintGroupbox);
    XML::writeAttribute(w, XML::Attrib::value, m_expandFootprint);
    w->writeEndElement();
}

void ScanItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            BeamItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // footprint
        } else if (tag == Tag::Footprint) {
            m_footprint.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // footprint groupbox: is expanded?
        } else if (tag == Tag::ExpandFootprintGroupbox) {
            XML::readAttribute(r, XML::Attrib::value, &m_expandFootprint);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

double ScanItem::getInclinationAngle() const
{
    return 0.0;
}

void ScanItem::setInclinationAngle(double value)
{
    ASSERT(value == 0.0);
    value = 0.0;
    BeamItem::setInclinationAngle(value);
}

SpecularBeamInclinationItem* ScanItem::inclinationAngleItem() const
{
    return dynamic_cast<SpecularBeamInclinationItem*>(BeamItem::inclinationAngleItem());
}

BasicAxisItem* ScanItem::inclinationAxisItem() const
{
    return inclinationAngleItem()->alphaAxisItem();
}

void ScanItem::setGaussianFootprint(double value)
{
    m_footprint.setCurrentItem(new FootprintGaussianItem(value));
}

void ScanItem::setSquareFootprint(double value)
{
    m_footprint.setCurrentItem(new FootprintSquareItem(value));
}

void ScanItem::updateToData(const IAxis& axis, QString units)
{
    if (units == "nbins") {
        inclinationAngleItem()->initUniformAxis(axis);
        inclinationAngleItem()->selectUniformAxis();
    } else {
        inclinationAngleItem()->initPointwiseAxis(axis, units);
        inclinationAngleItem()->selectPointwiseAxis();
    }
}

//  ************************************************************************************************
//  GISASBeamItem
//  ************************************************************************************************

GISASBeamItem::GISASBeamItem()
{
    m_wavelengthItem.reset(new BeamWavelengthItem);
    m_inclinationAngleItem.reset(new BeamInclinationAngleItem);
}

double GISASBeamItem::getInclinationAngle() const
{
    return m_inclinationAngleGetter
               ? m_inclinationAngleGetter()
               : dynamic_cast<BeamInclinationAngleItem*>(inclinationAngleItem())
                     ->inclinationAngle();
}

void GISASBeamItem::setInclinationAngleGetter(std::function<double()> getter)
{
    m_inclinationAngleGetter = getter;
}
