//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/BeamItems.h
//! @brief     Defines BeamItem hierarchy
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DEVICE_BEAMITEMS_H
#define BORNAGAIN_GUI_MODEL_DEVICE_BEAMITEMS_H

#include "GUI/Model/CatDevice/FootprintItemCatalog.h"
#include "GUI/Model/Descriptor/DoubleProperty.h"
#include "GUI/Model/Descriptor/SelectionProperty.h"
#include "GUI/Model/Device/BeamAngleItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/Model/Device/SpecularBeamInclinationItem.h"
#include <functional>
#include <heinz/Vectors3D.h>

class BasicAxisItem;
class Beam;
class BeamAzimuthalAngleItem;
class BeamWavelengthItem;
class FootprintItem;
class IAxis;

class BeamItem {
public:
    virtual ~BeamItem() = default;

    virtual void writeTo(QXmlStreamWriter* w) const;
    virtual void readFrom(QXmlStreamReader* r);

    DoubleProperty& intensity() { return m_intensity; }
    const DoubleProperty& intensity() const { return m_intensity; }
    void setIntensity(double v) { m_intensity.setValue(v); }

    double wavelength() const;
    void setWavelength(double value);
    BeamWavelengthItem* wavelengthItem() const;

    virtual double getInclinationAngle() const = 0;
    virtual void setInclinationAngle(double value);
    virtual BeamDistributionItem* inclinationAngleItem() const;

    double azimuthalAngle() const;
    void setAzimuthalAngle(double value);
    BeamAzimuthalAngleItem* azimuthalAngleItem() const;

    std::shared_ptr<Beam> createBeam() const;

    bool isExpandBeamParameters() const { return m_expandBeamParameters; }
    void setExpandBeamParameters(bool b) { m_expandBeamParameters = b; }

protected:
    BeamItem();

    std::unique_ptr<BeamWavelengthItem> m_wavelengthItem;
    std::unique_ptr<BeamAzimuthalAngleItem> m_azimuthalAngleItem;
    std::unique_ptr<BeamDistributionItem> m_inclinationAngleItem;

    DoubleProperty m_intensity;
    bool m_expandBeamParameters = true;
};

class ScanItem : public BeamItem {
public:
    explicit ScanItem(const InstrumentItem* owningInstrument);

    void writeTo(QXmlStreamWriter* w) const override;
    void readFrom(QXmlStreamReader* r) override;

    double getInclinationAngle() const override;
    void setInclinationAngle(double value) override;
    SpecularBeamInclinationItem* inclinationAngleItem() const override;
    BasicAxisItem* inclinationAxisItem() const;

    SelectionProperty<FootprintItemCatalog>& footprintSelection() { return m_footprint; }

    void setGaussianFootprint(double value);
    void setSquareFootprint(double value);

    void updateToData(const IAxis& axis, QString units);

    bool isExpandFootprint() const { return m_expandFootprint; }
    void setExpandFootprint(bool b) { m_expandFootprint = b; }

private:
    SelectionProperty<FootprintItemCatalog> m_footprint;
    bool m_expandFootprint = true;
    double m_wavelength = 0;
};

class GISASBeamItem : public BeamItem {
public:
    GISASBeamItem();

    double getInclinationAngle() const override;

    //! Explicitly set a getter function for the inclination angle.
    //!
    //! If a getter is set, it overrides the behavior in getInclinationAngle().
    //! Helpful if the angle cannot be determined from the internal beam parameters, but from
    //! external information.
    void setInclinationAngleGetter(std::function<double()> getter);

    std::function<double()> m_inclinationAngleGetter;
};

#endif // BORNAGAIN_GUI_MODEL_DEVICE_BEAMITEMS_H
